@@ -1,26 +1,28 @@
 # -*- coding: utf-8 -*-
 # pylint: disable=undefined-variable
+# type: ignore[name-defined]
+__version__ = '0.21.0'
+
 import logging
 
-from .loaders import *
 from .communications import *
 from .events import *
 from .exceptions import *
 from .futures import *
+from .loaders import *
+from .mixins import *
 from .persistence import *
 from .ports import *
-from .processes import *
-from .process_states import *
 from .process_comms import *
 from .process_listener import *
-from .mixins import *
+from .process_states import *
+from .processes import *
 from .utils import *
-from .version import *
 from .workchains import *
 
 __all__ = (
     events.__all__ + exceptions.__all__ + processes.__all__ + utils.__all__ + futures.__all__ + mixins.__all__ +
-    persistence.__all__ + communications.__all__ + process_comms.__all__ + version.__all__ + process_listener.__all__ +
+    persistence.__all__ + communications.__all__ + process_comms.__all__ + process_listener.__all__ +
     workchains.__all__ + loaders.__all__ + ports.__all__ + process_states.__all__
 )
 
